static char *rcsId = "$Header: /misc/hadaq/cvsroot/hadaq/nettrans.c,v 6.38 2005/05/02 13:58:36 hadaq Exp $";


#define _GNU_SOURCE
#include <unistd.h>
#include "grmblfx.h"

#include <assert.h>
#include <fcntl.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <syslog.h>

#include "hadtu.h"

#include "worker.h"
#include "nettrans.h"

static int min(int a, int b)
{
	return a < b ? a : b;
}

static int openUdp(NetTrans *my, unsigned long addr, int port, int fl)
{
	int retVal;

	assert(my != NULL);

	my->fd = socket(AF_INET, SOCK_DGRAM, 0);

	if (-1 == my->fd) {
		retVal = -1;
	} else if (fl == O_RDONLY) {
		int rcvBufLenReq = 1 * (1 << 20);
		int rcvBufLenRet;
		int rcvBufLenLen = sizeof(rcvBufLenReq);

		if (setsockopt(my->fd, SOL_SOCKET, SO_RCVBUF, &rcvBufLenReq, rcvBufLenLen) == -1) {
			syslog(LOG_WARNING, "setsockopt(..., SO_RCVBUF, ...): %s", strerror(errno));
		}

		if (getsockopt(my->fd, SOL_SOCKET, SO_RCVBUF, &rcvBufLenRet, &rcvBufLenLen) == -1) {
			syslog(LOG_WARNING, "getsockopt(..., SO_RCVBUF, ...): %s", strerror(errno));
		}
		if (rcvBufLenRet < rcvBufLenReq) {
			syslog(LOG_WARNING, "UDP receive buffer length (%d) smaller than requested buffer length (%d)", rcvBufLenRet,
				   rcvBufLenReq);
		}

		my->spec.sa.sin_family = AF_INET;
		my->spec.sa.sin_port = port;
		my->spec.sa.sin_addr.s_addr = htonl(INADDR_ANY);
		retVal = bind(my->fd, &my->spec.sa, sizeof(my->spec.sa));
	} else if (fl == O_WRONLY) {
		my->spec.sa.sin_family = AF_INET;
		my->spec.sa.sin_port = port;
		my->spec.sa.sin_addr.s_addr = addr;
	} else {
		retVal = -1;
	}
	my->mtuSize = 15 * 1024;

	if (0 > retVal) {
		syslog(LOG_DEBUG, "openUdp: %s", strerror(errno));
	}
	return retVal;
}

int openGeneric(NetTrans *my, const char *name, int flags, size_t bandwidth)
{
	int retVal;
	char fam[32];
	char pa[32];
	char ep[32];

	assert(my != NULL);
	assert(name != NULL);

	if (0 == (sscanf(name, "%[^:]:%[^:]:%s", fam, pa, ep))) {
		errno = EINVAL;
		retVal = -1;
	} else {
		if (strcmp(fam, "UDP") == 0) {
			unsigned long addr;
			int port;

			addr = inet_addr(pa);
			port = htons(atoi(ep));
			my->type = NetTransType_udp;
			retVal = openUdp(my, addr, port, flags);
		} else {
			my->type = NetTransType_invalid;
			errno = EINVAL;
			retVal = -1;
		}
	}

	if (0 > retVal) {
		syslog(LOG_DEBUG, "openGeneric: %s", strerror(errno));
	}
	return retVal;
}

int closeGeneric(NetTrans *my)
{
	int retVal;

	assert(my != NULL);

	switch (my->type) {
	case NetTransType_udp:
		close(my->fd);
		retVal = 0;
		break;
	}

	if (0 > retVal) {
		syslog(LOG_DEBUG, "closeGeneric: %s", strerror(errno));
	}
	return retVal;
}

int sendGeneric(NetTrans *my)
{
	int retVal;
	int socklen = sizeof(my->spec.sa);

	assert(NetTrans_invariant(my));

	switch (my->type) {
	case NetTransType_udp:
		for (retVal = sendto(my->fd, my->pkt, my->pktSize, 0, &my->spec.sa, socklen);
			 0 > retVal && errno == ENOBUFS; retVal = sendto(my->fd, my->pkt, my->pktSize, 0, &my->spec.sa, socklen)
			) {
			struct timespec tS, *t = &tS;
			t->tv_sec = 0;
			t->tv_nsec = 020000000;
			nanosleep(t, NULL);
		}
		break;
	}

	if (retVal != my->pktSize) {
		retVal = -1;
	}
	if (0 > retVal) {
		syslog(LOG_DEBUG, "sendGeneric: %s (%d)", strerror(errno), errno);
	}
	return retVal;
}

int recvGeneric(NetTrans *my)
{
	int retVal;
	int socklen = sizeof(my->spec.sa);

	assert(NetTrans_invariant(my));

	switch (my->type) {
	case NetTransType_udp:
		retVal = recvfrom(my->fd, my->pkt, my->mtuSize, 0, &my->spec.sa, &socklen);
		break;
	}

	if (0 > retVal) {
		syslog(LOG_DEBUG, "recvGeneric: %s", strerror(errno));
	} else {
		my->pktSize = retVal;
	}
	return retVal;
}

int NetTrans_invariant(const NetTrans *my)
{
	return (my != NULL && my->type == NetTransType_udp);
}

NetTrans *NetTrans_create(const char *name, size_t bandwidth, Worker *worker)
{
	NetTrans *my;
	char buf[80];
	static int instance = 0;

	my = malloc(sizeof(NetTrans));

	if (0 > openGeneric(my, name, O_RDONLY, bandwidth)) {
		return NULL;
	}
	my->pktSize = 0;
	my->offset = 0;

	my->pktsSent = NULL;
	my->msgsSent = NULL;
	my->bytesSent = NULL;
	sprintf(buf, "%s%d", "pktsReceived", instance);
	my->pktsReceived = Worker_addStatistic(worker, buf);
	sprintf(buf, "%s%d", "pktsDiscarded", instance);
	my->pktsDiscarded = Worker_addStatistic(worker, buf);
	sprintf(buf, "%s%d", "msgsReceived", instance);
	my->msgsReceived = Worker_addStatistic(worker, buf);
	sprintf(buf, "%s%d", "msgsDiscarded", instance);
	my->msgsDiscarded = Worker_addStatistic(worker, buf);
	sprintf(buf, "%s%d", "bytesReceived", instance);
	my->bytesReceived = Worker_addStatistic(worker, buf);

	instance++;

	return my;
}

NetTrans *NetTrans_open(const char *name, size_t bandwidth, Worker *worker)
{
	NetTrans *my;
	char buf[80];

	my = malloc(sizeof(NetTrans));

	if (0 > openGeneric(my, name, O_WRONLY, bandwidth)) {
		return NULL;
	}
	my->pktSize = 0;
	my->offset = 0;

	my->pktsSent = Worker_addStatistic(worker, "pktsSent");
	my->msgsSent = Worker_addStatistic(worker, "msgsSent");
	my->bytesSent = Worker_addStatistic(worker, "bytesSent");
	my->pktsReceived = NULL;
	my->pktsDiscarded = NULL;
	my->msgsReceived = NULL;
	my->msgsDiscarded = NULL;
	my->bytesReceived = NULL;

	return my;
}

void NetTrans_close(NetTrans *my)
{
	closeGeneric(my);
	free(my);
}

void NetTrans_remove(NetTrans *my)
{
	closeGeneric(my);
	free(my);
}

static size_t adjustedSize(NetTrans *my, void *hadTu)
{
	size_t size;

	size = HadTu_paddedSize(hadTu) + 32;
	return size;
}

int NetTrans_send(NetTrans *my, void *hadTu)
{
	char saveIt[32];
	size_t size;

	/* copy message header to end for check at receiver */
	memcpy(saveIt, HadTu_end(hadTu), 32);
	memcpy(HadTu_end(hadTu), hadTu, 32);

	size = adjustedSize(my, hadTu);

	my->offset = 0;
	do {
		my->pkt = (char *) hadTu + my->offset;
		my->pktSize = min(my->mtuSize, size - my->offset);

		if (0 > sendGeneric(my)) {
			return -1;
		}
		my->offset += my->pktSize;
		(*my->pktsSent)++;
	} while (my->pktSize == my->mtuSize);

	memcpy(HadTu_end(hadTu), saveIt, 32);

	(*my->msgsSent)++;
	(*my->bytesSent) += size;

	return 0;
}

static int assembleMsg(NetTrans *my, void *hadTu, size_t size)
{
	int retVal;

	/* BUGBUG quickhack for flushing the input data.  Properly in one of */
	/* the parent functions this should be handled, but this would need */
	/* better sunc. between nettrans & shmtrans */
	if (hadTu == NULL) {
		my->pkt = malloc(my->mtuSize);
		if (0 > recvGeneric(my)) {
			abort();
		}
		my->offset = 0;
		(*my->pktsDiscarded)++;
		free(my->pkt);
		return 0;
	}
	my->pkt = (char *) hadTu + my->offset;
	if (0 > recvGeneric(my)) {
		abort();
	}
	my->offset += my->pktSize;
	(*my->pktsReceived)++;

	if (my->pktSize == my->mtuSize) {
		retVal = 0;
	} else {
		size_t size;

		size = adjustedSize(my, hadTu);

		if (my->offset != size || memcmp(HadTu_end(hadTu), hadTu, 32)) {
			(*my->msgsDiscarded)++;
			retVal = 0;
		} else {
			(*my->msgsReceived)++;
			(*my->bytesReceived) += my->offset;
			retVal = 1;
		}
		my->offset = 0;
	}
	return retVal;
}

void NetTrans_recv(NetTrans *my, void *hadTu, size_t size)
{
	while (!assembleMsg(my, hadTu, size)) {
	}
}

unsigned long NetTrans_multiRecv(NetTrans *my[], void *hadTu[], size_t size[], int nrOfMsgs)
{
	int i;
	unsigned long msgIsComplete;
	fd_set fdSetS, *fdSet = &fdSetS;

	assert((1L << (nrOfMsgs - 1)) != 0);	/* is long long enough for result */

	msgIsComplete = 0;

	do {
		FD_ZERO(fdSet);
		for (i = 0; i < nrOfMsgs; i++) {
			FD_SET(my[i]->fd, fdSet);
		}
		if (0 > select(getdtablesize(), fdSet, NULL, NULL, NULL)) {
			abort();
		}
		for (i = 0; i < nrOfMsgs; i++) {
			if (FD_ISSET(my[i]->fd, fdSet)) {
				if (assembleMsg(my[i], hadTu[i], size[i])) {
					msgIsComplete |= 1 << i;
				}
			}
		}
	} while (0 && msgIsComplete == 0);

	return msgIsComplete;
}
